import pytest
from ..whooper import _s3 as s3


download_upload_TypeError_args = [
    (1, "foo"),
    ("foo", 1),
    (1, ["foo"]),
    (["foo"], 1),
    ([1], "foo"),
    ("foo", [1]),
    (["foo"], "bar"),
    ("foo", ["bar"]),
]


@pytest.mark.parametrize("s3_filepath,local_filepath", download_upload_TypeError_args)
def test_s3_download_type_error(s3_filepath, local_filepath):
    # test whether s3_download() raises the proper error
    with pytest.raises(TypeError):
        s3.s3_download(
            bucket="test", s3_filepath=s3_filepath, local_filepath=local_filepath
        )


@pytest.mark.parametrize("s3_filepath,local_filepath", download_upload_TypeError_args)
def test_s3_upload_type_error(s3_filepath, local_filepath):
    # test whether s3_upload() raises the proper error
    with pytest.raises(TypeError):
        s3.s3_upload(
            bucket="test", local_filepath=local_filepath, s3_filepath=s3_filepath
        )


download_upload_ValueError_args = [
    (["foo", "bar"], ["baz"]),
    (["foo"], ["bar", "baz"]),
    (["f*"], ["bar"]),
    (["foo"], ["b*"]),
]


@pytest.mark.parametrize("s3_filepath,local_filepath", download_upload_ValueError_args)
def test_s3_download_value_error(s3_filepath, local_filepath):
    # test whether s3_download() raises the proper error
    with pytest.raises(ValueError):
        s3.s3_download(
            bucket="test", s3_filepath=s3_filepath, local_filepath=local_filepath
        )


@pytest.mark.parametrize("s3_filepath,local_filepath", download_upload_ValueError_args)
def test_s3_upload_value_error(s3_filepath, local_filepath):
    # test whether s3_upload() raises the proper error
    with pytest.raises(ValueError):
        s3.s3_upload(
            bucket="test", local_filepath=local_filepath, s3_filepath=s3_filepath
        )


delete_TypeError_args = [(1,), ([1],)]


@pytest.mark.parametrize("s3_filepath", delete_TypeError_args)
def test_s3_delete_type_error(s3_filepath):
    # test whether s3_delete() raises the proper error
    with pytest.raises(TypeError):
        s3.s3_delete(bucket="test", s3_filepath=s3_filepath)


def test_s3_delete_value_error():
    # test whether s3_delete() raises the proper error
    with pytest.raises(ValueError):
        s3.s3_delete(bucket="test", s3_filepath=["*"])



# standard library imports
import os
from typing import Dict

# third party imports
import redshift_connector


def redshift_get_conn(env_var: str) -> redshift_connector.core.Connection:
    """Creates a Redshift connection object

    Parameters
    ----------
    env_var : str
        name of the environment variable containing the credentials str
        creds_str should have the below format where the user has inserted their values:
        'host=my_hostname database=my_database user=my_user password=my_password port=1234'

    Returns
    -------
    database connection object

    Example use
    -----------
    conn = redshift_get_conn(env_var='REDSHIFT_CREDS')
    """
    _env_var_validator(env_var=env_var)
    cred_str = os.environ[env_var]
    creds_dict = _create_creds_dict(cred_str)
    conn = redshift_connector.connect(**creds_dict)
    return conn


def read_sql(sql_filename: str) -> str:
    """Ingests a SQL file and returns a str containing the contents of the file

    Parameters
    ----------
    sql_filename : str
        path to and name of SQL file to be ingested

    Returns
    -------
    str
        contents of the SQL file ingested

    Example use
    -----------
    sql = read_sql(sql_filename='../sql/my_script.sql')
    """
    if not isinstance(sql_filename, str):
        raise TypeError("sql_filename must be of str type")
    with open(sql_filename, "r") as f:
        sql_str = " ".join(f.readlines())
    return sql_str


def redshift_execute_sql(
    sql: str, env_var: str, return_data: bool = False, return_dict: bool = False
):
    """Ingests a SQL query as a string and executes it (potentially returning data)

    Parameters
    ----------
    sql : str
        SQL query to be executed
    env_var : str
        name of the environment variable containing the credentials str
        creds_str should have the below format where the user has inserted their values:
        'host=my_hostname database=my_database user=my_user password=my_password port=1234'
    return_data : bool
        whether or not the query should return data
    return_dict : bool
        whether or not to return data as a dict (for easy ingestion into pandas)

    Returns
    -------
    None or list of str and list of tuples or dict
        if not return_data then None
        if return_data and not return_dict then list of str (column names) and list of tuples (data)
        if return_data and return_dict then dict with keys 'columns' and 'data' with values from above

    Example use
    -----------
    # Statement that does not return data (creating/dropping tables or copying/unloading data, etc)
    redshift_execute_sql(
        sql=sql,
        env_var='REDSHIFT_CREDS',
        return_data=False,
        return_dict=False)

    # Return data as a list of tuples and the columns as a list of str
    data, columns = redshift_execute_sql(
        sql=sql,
        env_var='REDSHIFT_CREDS',
        return_data=True,
        return_dict=False)

    # Return data for direct ingestion into pandas
    import pandas as pd
    df = pd.DataFrame(**redshift_execute_sql(
        sql=sql,
        env_var='REDSHIFT_CREDS',
        return_data=True,
        return_dict=True))
    """
    _redshift_execute_sql_arg_validator(
        sql=sql, env_var=env_var, return_data=return_data, return_dict=return_dict
    )
    with redshift_get_conn(env_var=env_var) as conn:
        with conn.cursor() as cursor:
            cursor.execute(sql)
            if return_data:
                columns = [desc[0] for desc in cursor.description]
                data = [row for row in cursor]
                conn.commit()
                if return_dict:
                    return {"data": data, "columns": columns}
                else:
                    return data, columns
            else:
                conn.commit()
                return


def _create_creds_dict(creds_str: str) -> Dict:
    """Takes the credentials str and converts it to a dict

    Parameters
    ----------
    creds_str : str
        credentials string with the below format where the user has inserted their values:
        'host=my_hostname database=my_database user=my_user password=my_password port=1234'

    Returns
    -------
    dict
        credentials in dict form
    """
    creds_dict = {}
    for param in creds_str.split(" "):
        split_param = param.split("=")
        if split_param[0] == "port":
            split_param[1] = int(split_param[1])
        creds_dict[split_param[0]] = split_param[1]
    return creds_dict


def _env_var_validator(env_var: str) -> None:
    """Validates that the user is providing the an environment variable name, rather than the credentials string

    Parameters
    ----------
    env_var : str
        the name of the environment variable referencing the Redshift credential string

    Returns
    -------
    None
    """
    creds_str_keys = ["host", "database", "user", "password", "port"]
    if all(key in env_var for key in creds_str_keys):
        raise ValueError(
            "This field should contain the name of an env variable, not the credentials string."
        )
    elif env_var not in os.environ:
        raise KeyError("Redshift credentials env variable not found.")
    return


def _redshift_execute_sql_arg_validator(
    sql: str,
    env_var: str,
    return_data: bool,
    return_dict: bool,
) -> None:
    """Validates the redshift_execute_sql arguments and raises clear errors

    Parameters
    ----------
    sql : str
        SQL query to be executed
    env_var : str
        name of the environment variable containing the credentials str
    return_data : bool
        whether or not the query should return data
    return_dict : bool
        whether or not to return data as a dict (for easy ingestion into pandas)

    Returns
    -------
    None
    """
    for arg in [sql, env_var]:
        if not isinstance(arg, str):
            raise TypeError("sql and env_var must be of str type")
    for arg in [return_data, return_dict]:
        if not isinstance(arg, bool):
            raise TypeError("return_data and return_dict must be of bool type")
    return



# TODO: write boto_get_creds() tests (types, errors, and str contents)
import boto3
from ..whooper import _boto as bt


def test_boto_create_session_type():
    # test whether _create_session() returns the proper type
    assert isinstance(bt.boto_create_session(), boto3.session.Session)



"""Convenience wrappers for connecting to AWS S3 and Redshift"""

__version__ = "0.0.0"


# Boto3 function
from ._boto import boto_get_creds
from ._boto import boto_create_session

# Redshift functions
from ._redshift import redshift_get_conn
from ._redshift import read_sql
from ._redshift import redshift_execute_sql

# S3 functions
from ._s3 import s3_get_bucket
from ._s3 import s3_download
from ._s3 import s3_upload
from ._s3 import s3_delete


__all__ = ["_boto", "_redshift", "_s3"]



# standard library imports
from typing import Optional

# third party imports
import boto3


def boto_create_session(
    profile_name: str = "default",
    region_name: str = "us-west-2",
) -> boto3.session.Session:
    """Instantiates and returns a boto3 session object

    Parameters
    ----------
    profile_name : str
        profile name under which credentials are stored (default 'default' unless organization specific)
    region_name : str
        name of AWS regions (default 'us-west-2')

    Returns
    -------
    boto3 session object

    Example use
    -----------
    session = create_session(profile_name='default', region_name='us-west-2')
    """
    return boto3.session.Session(profile_name=profile_name, region_name=region_name)


def boto_get_creds(
    profile_name: str = "default",
    region_name: str = "us-west-2",
    session: Optional = None,
) -> str:
    """Generates and returns an S3 credential string

    Parameters
    ----------
    profile_name : str
        profile name under which credentials are stores (default 'default' unless organization specific)
    region_name : str
        name of AWS regions (default 'us-west-2')
    session : boto3 session object or None
        you can optionally provide a boto3 session object or the function can instantiate a new one if None

    Returns
    -------
    str
        credentials for accessing S3

    Example use
    -----------
    creds = boto_get_creds(
        profile_name='default',
        region_name='us-west-2',
        session=None)
    """
    if session is None:
        session = boto_create_session(
            profile_name=profile_name, region_name=region_name
        )
    access_key = session.get_credentials().access_key
    secret_key = session.get_credentials().secret_key
    token = session.get_credentials().token
    return f"""aws_access_key_id={access_key};aws_secret_access_key={secret_key};token={token}"""



import os
import pytest
from ..whooper import _redshift as rs


os.environ[
    "TEST_CREDS"
] = "host=my_hostname database=database user=my_user password=my_password port=1234"


def test_redshift_get_conn_value_error():
    # test whether redshift_get_conn() raises a ValueError when passed a connection str
    with pytest.raises(ValueError):
        rs.redshift_get_conn(os.environ["TEST_CREDS"])


def test_read_sql_type_error():
    # test whether read_sql() raises the proper error
    with pytest.raises(TypeError):
        rs.read_sql(1)


def test_read_sql_return_type():
    # test type returned by read_sql()
    assert isinstance(rs.read_sql("test/test.sql"), str)


def test_read_sql_contents():
    # test whether contents of str returned by read_sql() are correct
    test_str = "select\n     col1\n     col2\n from\n     pretend.first_table\n limit\n     1000;"
    assert rs.read_sql("test/test.sql") == test_str


redshift_execute_sql_TypeError_args = [
    (1, "foo", True, True),
    ("foo", 1, True, True),
    (1, 1, True, True),
    ("foo", "bar", True, "True"),
    ("foo", "bar", "True", True),
    ("foo", "bar", True, 1),
    ("foo", "bar", 1, True),
]


@pytest.mark.parametrize(
    "sql,env_var,return_data,return_dict", redshift_execute_sql_TypeError_args
)
def test_redshift_execute_sql_type_error(sql, env_var, return_data, return_dict):
    # test whether redshift_execute_sql() raises the proper error
    with pytest.raises(TypeError):
        rs.redshift_execute_sql(
            sql=sql, env_var=env_var, return_data=return_data, return_dict=return_dict
        )


def test_create_creds_dict_type():
    # test type returned by _create_creds_dict()
    assert isinstance(rs._create_creds_dict(os.environ["TEST_CREDS"]), dict)


def test_create_creds_dict_len():
    # test length of dict returned by _create_creds_dict()
    assert len(rs._create_creds_dict(os.environ["TEST_CREDS"])) == 5


def test_create_creds_dict_keys():
    # test whether proper keys are present in dict returned by _create_creds_dict()
    keys = ["host", "database", "user", "password", "port"]
    creds_dict = rs._create_creds_dict(os.environ["TEST_CREDS"])
    assert all(key in creds_dict for key in keys)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['whooper']

package_data = \
{'': ['*']}

install_requires = \
['boto3 >=1.17.1', 'redshift-connector >=2.0.873']

setup(name='whooper',
      version='0.0.0',
      description='Convenience wrappers for connecting to AWS S3 and Redshift',
      author='Nick Buker',
      author_email='nickbuker@gmail.com',
      url='https://github.com/nickbuker/whooper',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.6,<4',
     )



# standard library imports
import glob
import os
from typing import List

# third party imports
from ._boto import boto_create_session
import boto3
from boto3.exceptions import S3UploadFailedError
from boto3.s3.transfer import TransferConfig
from botocore.exceptions import ClientError


def s3_get_bucket(
    bucket: str,
    profile_name: str = "default",
    region_name: str = "us-west-2",
) -> boto3.resource("s3").Bucket:
    """Creates and returns a boto3 bucket object

    Parameters
    ----------
    bucket : str
        name of S3 bucket
    profile_name: str
        profile name for credentials (default 'default' or organization-specific)
    region_name : str
        name of AWS regions (default 'us-west-2')

    Returns
    -------
    boto3 bucket object

    Example use
    -----------
    bucket = s3_get_bucket(
        bucket='my_bucket',
        profile_name='default',
        region_name='us-west-2')
    """
    session = boto_create_session(profile_name=profile_name, region_name=region_name)
    s3 = session.resource("s3")
    my_bucket = s3.Bucket(bucket)
    try:
        s3.meta.client.head_bucket(Bucket=bucket)
    except ClientError as e:
        # Check if bucket exists, if not raise error
        error_code = int(e.response["Error"]["Code"])
        if error_code == 404:
            raise NameError("404 Bucket does not exist")
        if error_code == 400:
            raise NameError("400 The credentials were expired or incorrect.")
    return my_bucket


def s3_download(
    bucket: str,
    s3_filepath: str,
    local_filepath: str,
    profile_name: str = "default",
    region_name: str = "us-west-2",
    multipart_threshold: int = 8388608,
    multipart_chunksize: int = 8388608,
) -> None:
    """Downloads a file or collection of files from S3

    Parameters
    ----------
    bucket : str
        name of S3 bucket
    s3_filepath : str or list
        path and filename within bucket to file(s) you would like to download
    local_filepath : str or list
        path and filename for file(s) to be saved locally
    profile_name : str
        profile name for credentials (default 'default' or organization-specific)
    region_name : str
        name of AWS region (default value 'us-west-2')
    multipart_threshold : int
        minimum file size to initiate multipart download
    multipart_chunksize : int
        chunksize for multipart download

    Returns
    -------
    None

    Example use
    -----------
    # Downloading a single file from S3:
    s3_download(
        bucket='my_bucket',
        s3_filepath='tmp/my_file.csv',
        local_filepath='../data/my_file.csv')

    # Downloading with a profile name:
    s3_download(
        bucket='my_bucket',
        profile_name='my-profile-name',
        s3_filepath='tmp/my_file.csv',
        local_filepath='../data/my_file.csv')

    # Downloading a list of files from S3 (will not upload contents of subdirectories):
    s3_download(
        bucket='my_bucket',
        s3_filepath=['tmp/my_file1.csv', 'tmp/my_file2.csv', 'img.png'],
        local_filepath=['../data/my_file1.csv', '../data/my_file2.csv', '../img.png'])

    # Downloading files matching a pattern from S3 (will not upload contents of subdirectories):
    s3_download(
        bucket='my_bucket',
        s3_filepath='tmp/*.csv',
        local_filepath='../data/')

    # Downloading all files in a directory from S3 (will not upload contents of subdirectories):
    s3_download(
        bucket='my_bucket',
        s3_filepath='tmp/*',
        local_filepath='../data/')
    """
    # validate s3_filepath and local_filepath arguments
    _download_upload_filepath_validator(
        s3_filepath=s3_filepath, local_filepath=local_filepath
    )
    # create bucket object
    my_bucket = s3_get_bucket(
        bucket=bucket, profile_name=profile_name, region_name=region_name
    )
    # multipart_threshold and multipart_chunksize, defaults = Amazon defaults
    config = TransferConfig(
        multipart_threshold=multipart_threshold, multipart_chunksize=multipart_chunksize
    )
    if isinstance(s3_filepath, str):
        # find keys matching wildcard
        if "*" in s3_filepath:
            s3_filepath = _s3_glob(s3_filepath=s3_filepath, my_bucket=my_bucket)
            local_filepath = [
                os.path.join(local_filepath, key.split("/")[-1]) for key in s3_filepath
            ]
        # insert into list so same looping structure can be used
        else:
            s3_filepath = [s3_filepath]
            local_filepath = [local_filepath]
    # download all files from S3
    for s3_key, local_file in zip(s3_filepath, local_filepath):
        try:
            my_bucket.download_file(s3_key, local_file, Config=config)
        except ClientError as e:
            error_code = int(e.response["Error"]["Code"])
            if error_code == 400:
                raise NameError("The credentials are expired or not valid. " + str(e))
            else:
                raise e
    return


def s3_upload(
    bucket: str,
    local_filepath: str,
    s3_filepath: str,
    profile_name: str = "default",
    region_name: str = "us-west-2",
    multipart_threshold: int = 8388608,
    multipart_chunksize: int = 8388608,
) -> None:
    """Uploads a file or collection of files to S3

    Parameters
    ----------
    bucket : str
        name of S3 bucket
    local_filepath : str or list
        path and filename(s) to be uploaded
    s3_filepath : str or list
        path and filename(s) within the bucket for the file to be uploaded
    region_name : str
        name of AWS region (default value 'us-west-2')
    profile_name : str
        profile name for credentials (default 'default' or organization-specific)
    multipart_threshold : int
        minimum file size to initiate multipart upload
    multipart_chunksize : int
        chunksize for multipart upload

    Returns
    -------
    None

    Example use
    -----------

    # Uploading a single file to S3:
    s3_upload(
        bucket='my_bucket',
        local_filepath='../data/my_file.csv',
        s3_filepath='tmp/my_file.csv')

    # Uploading with a profile name:
    s3_upload(
        bucket='my_bucket',
        profile_name='my-profile-name',
        local_filepath='../data/my_file.csv',
        s3_filepath='tmp/my_file.csv')

    Uploading a list of files to S3 (will not upload contents of subdirectories):
    s3_upload(
        bucket='my_bucket',
        local_filepath=['../data/my_file1.csv', '../data/my_file2.csv', '../img.png'],
        s3_filepath=['tmp/my_file1.csv', 'tmp/my_file2.csv', 'img.png'])

    Uploading files matching a pattern to S3 (will not upload contents of subdirectories):
    s3_upload(
        bucket='my_bucket',
        local_filepath='../data/*.csv',
        s3_filepath='tmp/')

    Uploading all files in a directory to S3 (will not upload contents of subdirectories):
    s3_upload(
        bucket='my_bucket',
        local_filepath='../data/*'
        s3_filepath='tmp/')
    """
    _download_upload_filepath_validator(
        s3_filepath=s3_filepath, local_filepath=local_filepath
    )
    my_bucket = s3_get_bucket(
        bucket=bucket, profile_name=profile_name, region_name=region_name
    )
    # multipart_threshold and multipart_chunksize, defaults = Amazon defaults
    config = TransferConfig(
        multipart_threshold=multipart_threshold, multipart_chunksize=multipart_chunksize
    )
    if isinstance(local_filepath, str):
        if "*" in local_filepath:
            items = glob.glob(local_filepath)
            # filter out directories
            local_filepath = [item for item in items if os.path.isfile(item)]
            tmp_s3_filepath = [s3_filepath + f.split("/")[-1] for f in local_filepath]
            s3_filepath = tmp_s3_filepath
        else:
            local_filepath = [local_filepath]
            s3_filepath = [s3_filepath]
    # upload all files to S3
    for local_file, s3_key in zip(local_filepath, s3_filepath):
        try:
            my_bucket.upload_file(local_file, s3_key, Config=config)
        except boto3.exceptions.S3UploadFailedError as e:
            raise S3UploadFailedError(str(e))
    return


def s3_delete(
    bucket: str, s3_filepath: str, profile_name="default", region_name="us-west-2"
) -> List[str]:
    """Deletes a file or collection of files from S3

    Parameters
    ----------
    bucket : str
        name of S3 bucket
    s3_filepath : str or list
        path and filename of item(s) within the bucket to be deleted
    profile_name : str
        profile name for credentials (default 'default' or organization-specific)
    region_name : str
        name of AWS region (default value 'us-west-2')

    Returns
    -------
    List
        Deleted keys

    Example use
    -----------
    # Deleting a single file in S3:
    resp = s3_delete(bucket='my_bucket', s3_filepath='tmp/my_file.csv')

    # Deleting with a profile name:
    s3_upload(
        bucket='my_bucket',
        profile_name='my-profile-name',
        s3_filepath='tmp/my_file.csv')

    # Deleting a list of files in S3:
    resp = s3_delete(
        bucket='my_bucket',
        s3_filepath=['tmp/my_file1.csv', 'tmp/my_file2.csv', 'img.png'])

    # Deleting files matching a pattern in S3:
    resp = s3_delete(bucket='my_bucket', s3_filepath='tmp/*.csv')

    # Deleting all files in a directory in S3:
    resp = s3_delete(bucket='my_bucket', s3_filepath='tmp/*')
    """
    _delete_filepath_validator(s3_filepath=s3_filepath)
    my_bucket = s3_get_bucket(
        bucket=bucket, profile_name=profile_name, region_name=region_name
    )
    if isinstance(s3_filepath, str):
        if "*" in s3_filepath:
            s3_filepath = _s3_glob(s3_filepath=s3_filepath, my_bucket=my_bucket)
            if not s3_filepath:  # check if this list of S3 filepaths is empty
                return []
        else:
            s3_filepath = [s3_filepath]
    del_dict = {}
    objects = []
    for key in s3_filepath:
        objects.append({"Key": key})
    del_dict["Objects"] = objects
    response = my_bucket.delete_objects(Delete=del_dict)
    return response["Deleted"]


def _download_upload_filepath_validator(
    s3_filepath: str,
    local_filepath: str,
) -> None:
    """Validates the s3_filepath and local_filepath arguments and raises clear errors

    Parameters
    ----------
    s3_filepath : str or list of str
        path and filename of item(s) within the S3 bucket
    local_filepath : str or list of str
        path and filename for local file(s)

    Returns
    -------
    None
    """
    for arg in (s3_filepath, local_filepath):
        if not isinstance(arg, (list, str)):
            raise TypeError(
                "Both s3_filepath and local_filepath must be of type list or str"
            )
    if type(s3_filepath) != type(local_filepath):
        raise TypeError("Both s3_filepath and local_filepath must be of same type")
    if isinstance(s3_filepath, list):
        for f in s3_filepath + local_filepath:
            if not isinstance(f, str):
                raise TypeError(
                    "If s3_filepath and local_filepath are lists, they must contain strings"
                )
            if "*" in f:
                raise ValueError(
                    "Wildcards (*) are not permitted within a list of filepaths"
                )
        if len(s3_filepath) != len(local_filepath):
            raise ValueError(
                "The s3_filepath list must the same number of elements as the local_filepath list"
            )
    return


def _delete_filepath_validator(s3_filepath: str) -> None:
    """Validates the s3_filepath argument and raises clear errors

    Parameters
    ----------
    s3_filepath : str or list of str
        path and filename of item(s) within the S3 bucket

    Returns
    -------
    None
    """
    if not isinstance(s3_filepath, (list, str)):
        raise TypeError("s3_filepath must be of type list or str")
    if isinstance(s3_filepath, list):
        for f in s3_filepath:
            if not isinstance(f, str):
                raise TypeError("If s3_filepath is a list, it must contain strings")
            if "*" in f:
                raise ValueError(
                    "Wildcards (*) are not permitted within a list of filepaths"
                )
    return


def _s3_glob(
    s3_filepath: str,
    my_bucket: str,
) -> List[str]:
    """Searches a directory in an S3 bucket and returns keys matching the wildcard

    Parameters
    ----------
    s3_filepath : str
        the S3 filepath (with wildcard) to be searched for matches
    my_bucket : boto3 bucket object
        the S3 bucket object containing the directories to be searched

    Returns
    -------
    list of str
        S3 filepaths matching the wildcard
    """
    # use left and right for pattern matching
    left, _, right = s3_filepath.partition("*")
    # construct s3_path without wildcard
    s3_path = "/".join(s3_filepath.split("/")[:-1]) + "/"
    filtered_s3_filepath = []
    for item in my_bucket.objects.filter(Prefix=s3_path):
        # filter out directories
        if item.key[-1] != "/":
            p1, p2, p3 = item.key.partition(left)
            # pattern matching
            if p1 == "" and p2 == left and right in p3:
                filtered_s3_filepath.append(item.key)
    return filtered_s3_filepath
